%**********************************************************************;
% Project           : Iterative methods and preconditioning for large 
%			and sparse linear systems with applications
%
% Program name      : fgmres.m
%
% Author            : F. Durastante -- fdurastante@uninsubria.it
%
% Date created      : 28 July 2017
%
% Purpose           : Flexible GMRES with inner GMRES(m) prec. for Ax = b
%
% Revision History  :
%
% Date        Author      	Ref    Revision (Date in DD/MM/YYYY format) 
% 28/07/2017  F. Durastante     1      File Created
%
%**********************************************************************;
function [x,res,its] = fgmres(A,b,restart,maxit,tol,RESTART,TOL,MAXIT)
%%FGMRES+GMRES(m) Flexible-GMRES algorithm with restart preconditioned with
% GMRES(m) algorithm (using the MATLAB/OCTAVE implementation).
% It is based on restarted GMRES with Krylov subspace of dim = restart.

machine_tolerance = eps;
n = size(A,1);
its = 0;                  
x = sparse(n,1);

res = -ones(maxit,1);
c = zeros(maxit,1);
s = zeros(maxit,1);
rs = zeros(maxit,1);
while (its < maxit)
    vv(1:n,1) = b - A*x;
    ro = norm(vv(1:n,1),2);
    res(its+1) = ro;
    if (its  == 0)
        tol1=tol*ro;
    end
    if (ro <= tol1 || its >= maxit)
        return
    end
    t = 1.0/ ro;
    vv(1:n,1) = vv(1:n,1) * t;
    rs(1) = ro;
    if (verbose)
        fprintf(1,' its %d  res %e \n',its,ro)
    end
    i = 0;
    while (i < restart  &&  (ro  >  tol1)  &&  its < maxit)
        i=i+1;
        its = its + 1;
        i1 = i + 1;
	% To apply FGMRES with other variable preconditioners
	% change this line with the relative function handle
	% or external function, also you can include ILU or other
	% preconditioner in the internal GMRES
        [z,~] = gmres(A,vv(:,i),RESTART,TOL,MAXIT);
        Z(:,i) = z;
        vv(1:n,i1) = A*z;
        for j=1:i
            t = vv(1:n,j)'*vv(1:n,i1);
            hh(j,i) = t;
            vv(1:n,i1) = vv(1:n,i1) - t*vv(1:n,j);
        end
        t = norm(vv(1:n,i1),2);
        hh(i1,i) = t;
        if (t  ~= 0.0)
            t = 1.0 / t;
            vv(1:n,i1) = vv(1:n,i1)*t;
        end
        if (i ~= 1)
            for k=2:i
                k1 = k-1;
                t = hh(k1,i);
                hh(k1,i) = c(k1)*t + s(k1)*hh(k,i);
                hh(k,i) = -s(k1)*t + c(k1)*hh(k,i);
            end
        end
        den = sqrt(hh(i,i)^2 + hh(i1,i)^2);
        if (den  == 0.0)		% Fault Tolerance for
            den = machine_tolerance;	% Givens Rotations
        end 
        c(i) = hh(i,i)/den;
        s(i) = hh(i1,i)/den;
        rs(i1) = -s(i)*rs(i);
        rs(i) =  c(i)*rs(i);
        hh(i,i) = c(i)*hh(i,i) + s(i)*hh(i1,i);
        ro = abs(rs(i1));
        res(its+1) = ro;
    end
    rs(i) = rs(i)/hh(i,i);
    for  k=i-1:-1:1
        t=rs(k);
        for j=k+1:i
            t = t-hh(k,j)*rs(j);
        end  
        rs(k) = t/hh(k,k);
    end  
    for j=1:i
        x = x +rs(j)*Z(1:n,j);
    end
    if ((ro  <=  tol1) || (its >= maxit))
        res(res==-1)=[];
        return 
    end
end

end
