%**********************************************************************;
% Project           : Iterative methods and preconditioning for large 
%			and sparse linear systems with applications
%
% Program name      : jacobi.m
%
% Author            : F. Durastante -- fdurastante@uninsubria.it
%
% Date created      : 28 July 2017
%
% Purpose           : Jacobi method for the solutions of diagonally dominant Ax = b
%
% Revision History  :
%
% Date        Author      	Ref    Revision (Date in DD/MM/YYYY format) 
% 28/07/2017  F. Durastante     1      File Created
%
%**********************************************************************;
function [ x,varargout ] = jacobi(A,b,varargin)
%JACOBI method for determining the solutions of a diagonally dominant system 
%of linear equations.
%   A Matrix of the linear system
%   b Right-hand side
%   tol Solution tolerance, if not included tol = 1e-6
%   maxit Maximum number of iteration, if not included max(size(A,1),1000);
%   x Initial Guess (Input) Solution (Output), if not included x = [0,0,...,0];
%   flag (output) if 0 system solved
%   res Latest residual
%   iter Number of iteration
%   residual Convergence history

nargoutchk(0,5);
narginchk(0,5);

% Checking the data
if isempty(A)
    error('The matrix A has to be not empty');
end
if isempty(b)
    error('The vector b has to be not empty');
end
[n,m] = size(A);
if n ~= m
    error('A is not square');
end
if all(size(b) ~= [n,1])
    if all(size(b) == [1,n])
        b = b';
    else
        error('b dimension mismatch.')
    end
end

nVarargs = length(varargin);
if isempty(varargin)
    tol = norm(b,2)*1e-6;
    maxit = min(n,1000);
    x = zeros(n,1);
elseif nVarargs == 1
    if isnumeric(varargin{1}) && length(varargin{1}) == 1 ...
            && varargin{1} < 1
        tol = varargin{1};
        maxit = min(n,1000);
        x = zeros(n,1);
    else
        error('Tolerance not acceptable');
    end
elseif nVarargs == 2
    if isnumeric(varargin{1}) && length(varargin{1}) == 1 ...
            && varargin{1} < 1
        tol = varargin{1};
    else
        error('Tolerance not acceptable');
    end
    if isnumeric(varargin{2}) && length(varargin{2}) == 1 ...
            && varargin{2} > 0
        maxit = floor(varargin{2});
        x = zeros(n,1);
    else
        error('maxit not acceptable');
    end
elseif nVarargs == 3
    if isnumeric(varargin{1}) && length(varargin{1}) == 1 ...
            && varargin{1} < 1
        tol = varargin{1};
    else
        error('Tolerance not acceptable');
    end
    if isnumeric(varargin{2}) && length(varargin{2}) == 1
        maxit = floor(varargin{2});
    else
        error('maxit not acceptable');
    end
    if isnumeric(varargin{3}) && all(size(varargin{3}) == [n,1])
        x = floor(varargin{3});
    elseif isnumeric(varargin{3}) && all(size(varargin{3}) == [1,n])
        x = floor(varargin{3})';
    else
        error('initial data not acceptable');
    end
end

% Initialize residual if requested
nArgout = max(nargout,1) - 1;
if nArgout == 4
    residual = zeros(maxit,1);
end

% Doing the computation
D = diag(A);
bnorm = norm(b);
tol = tol*bnorm;
for i=1:maxit
    x = x + (b - A*x)./D;
    res = norm(b-A*x);
    %fprintf('%d & %2.2e\n',i,res);
    if nArgout == 4
        residual(i) = res;
    end
    if res < tol
        if nArgout == 1
            varargout{1} = 1;
            return
        elseif nArgout == 2
            varargout{1} = 1;
            varargout{2} = res;
            return
        elseif nArgout == 3;
            varargout{1} = 1;
            varargout{2} = res;
            varargout{3} = i;
            return
        elseif nArgout == 4;
            varargout{1} = 1;
            varargout{2} = res;
            varargout{3} = i;
            residual(residual == 0) = [];
            varargout{4} = residual;
            return
        end
    end
end

if nArgout == 1
    varargout{1} = 0;
    return
elseif nArgout == 2
    varargout{1} = 0;
    varargout{2} = res;
    return
elseif nArgout == 3;
    varargout{1} = 0;
    varargout{2} = res;
    varargout{3} = maxit;
    return
elseif nArgout == 4;
    varargout{1} = 0;
    varargout{2} = res;
    varargout{3} = maxit;
    residual(residual == 0) = [];
    varargout{4} = residual;
    return
end


end

