// -*- c++ -*-
/*
jacobi_host - Example: Jacobi Method over the CPU

Copyright (C) 2017

Fabio Durastante

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

// Standard C Library
#include <stdio.h> 

// CUSP LIBRARY
#include <cusp/array1d.h>
#include <cusp/csr_matrix.h>
#include <cusp/monitor.h>

#include <cusp/blas/blas.h>
#include <cusp/linear_operator.h>
#include <cusp/gallery/poisson.h>

// Include CUSP Jacobi Header
#include <cusp/relaxation/jacobi.h>

int main(int argc, char const *argv[])
{
  if(argc != 5){
    fprintf(stderr, "Correct usage is: %s n m tol maxit\n",argv[0] );
    return 1;
  }
  
  printf("\nExecuting %s Program\n", argv[0]);

  int n = atoi(argv[1]);  // Number of x grid points
  int m = atoi(argv[2]);  // Number of y grid points
  float tol = atof(argv[3]);  // Tolerance for the method
  int maxit = atoi(argv[4]);  // Maximum number of iteration
  

  // Change it to cusp::host_memory to run on the CPU
  typedef cusp::device_memory MemorySpace;


  // Measuring Time on the GPU
  cudaEvent_t start, stop;
  cudaEventCreate(&start);
  cudaEventCreate(&stop); 

  cusp::csr_matrix<int, float, MemorySpace> A; // Construct 5-pt Poisson example
  cusp::gallery::poisson5pt(A, n, m);
  cusp::array1d<float, MemorySpace> x(A.num_rows, 0); // Allocate/Initialize Memory
  cusp::array1d<float, MemorySpace> b(A.num_rows, 1); // b = [1,1,...,1]^T
  cusp::array1d<float, MemorySpace> r(A.num_rows); // Temporary data for the residual
  cusp::relaxation::jacobi<float, MemorySpace> M(A); // Construct Jacobi Method
  cusp::multiply(A, x, r); // Compute initial residual
  cusp::blas::axpy(b, r, float(-1));
  cusp::monitor<float> monitor(b, maxit, tol, 0, false); // Convergence Monitor
  
  cudaEventRecord(start); //tic
  while (!monitor.finished(r)) // Jacobi Iteration
  {
      M(A, b, x);
      cusp::multiply(A, x, r); // r = A*x
      cusp::blas::axpy(b, r, float(-1)); //r = r + (-1)*b
      ++monitor; // Increase the Iteration by one
  }
  cudaEventRecord(stop); //toc
  
  cudaEventSynchronize(stop);
  float milliseconds = 0;
  cudaEventElapsedTime(&milliseconds, start, stop);
  milliseconds *= 1e-3; // 1 ms = 1e-3 s

  printf("\n%d & %d & %2.2e & %2.2e \\\\\n",n*m,(int) monitor.iteration_count(),monitor.residual_norm(),milliseconds);

  return 0;
}