%% KRYLOV METHODS BASED ON OBLIQUE PROJECTION
% Test of oblique projection methods for the not Hermitian case obtained
% from the finite difference discretization of the advection-diffusion
% equation:
% -k div(grad(u))-nu grad(u) = f, for (x,y) \in [0,1]x[0,1]
%  u(x,0) = u(x,1) = u(1,y) = u(0,y) = 0 
% the code permits the usage of different number of grid points in the two
% side of the discretizations.
% ----------------------------------------------------------------------- %
clear;
clc;
% ----------------------------------------------------------------------- %

N = input('Number of grid points on the x axis: ');
M = input('Number of grid points on the y axis: ');
K = input('Diffusion Coefficient: ');
nu = input('Convection Velocity: ');

f = @(x,y) sin(x).*cos(y);

x = linspace(0,1,N);
y = linspace(0,1,M);
[X,Y] = meshgrid(x,y);
eN = ones(N,1); hx = 1/(N-1);
eM = ones(M,1); hy = 1/(M-1);
Tx = spdiags([-eN,2*eN,-eN]./hx,-1:1,N,N);
Ty = spdiags([-eM,2*eM,-eM]./hy,-1:1,M,M);
Bx = spdiags([-eN,eN]./(2*hx),[-1,1],N,N);
By = spdiags([-eM,eM]./(2*hy),[-1,1],M,M);

A = kron(K*Tx + nu*Bx,speye(M))+ kron(speye(N),K*Ty + nu*By);
b = reshape(f(X,Y),N*M,1);
n = N*M;

tol = 1e-6;

%% BiCGstab
tic; [x0,~,~,iter0,res0] = bicgstab(A,b,tol,n); time0 = toc;

%% BiCGstab(ell)
OPTIONS.Tol = tol;        
OPTIONS.MaxIt = n;       
OPTIONS.Disp = 0;
OPTIONS.ell = 2;
tic; [x1,hist1] = cgstab(A,b,[],[],OPTIONS); time1 = toc;

%% QMR
tic; [x3,~,~,iter3,res3] = qmr(A,b,tol,n); time3 = toc;

%% GMRES
tic; [x4,~,~,iter4,res4] = gmres(A,b,[],tol,n); time4 = toc;

%% GMRES(m)
tic; [x5,~,~,iter5,res5] = gmres(A,b,20,tol,n); time5 = toc;

%% BiCG
tic; [x6,~,~,iter6,res6] = bicg(A,b,tol,n); time6 = toc;

%% CGS
tic; [x7,~,~,iter7,res7] = cgs(A,b,tol,n); time7 = toc;

%% Prepare for the PLOT by transforming iteration into number of matvec
res0=repmat(res0,1,2)';     % 2 MatVec x Iteration BiCGstab
res0=res0(:)';
hist1=repmat(hist1(:,1),1,4)';   % 4 MatVec x Iteration BiCGstab(2)
hist1=hist1(:)';
res3=repmat(res3,1,2)';     % 2 MatVec x Iteration BiCGstab
res3=res3(:)';
res6=repmat(res6,1,2)';     % 2 MatVec x Iteration BiCG
res6=res6(:)';
res7=repmat(res7,1,2)';     % 2 MatVec x Iteration CGS
res7=res7(:)';

%% PLOT
figure(1)
loglog(1:2:length(res0),res0(1:2:end),'ko-',...
    1:2:length(res3),res3(1:2:end),'ks-',...
    1:4:length(hist1),hist1(1:4:end),'kd-',...
    1:2:length(res6),res6(1:2:end),'k*-',...
    1:2:length(res7),res7(1:2:end),'k+-');
legend('BiCGstab','QMR','BiCGstab(2)','BiCG','CGS')
xlabel('Number of MatVec operations');
ylabel('Norm of the Residual');

figure(2)
semilogy(1:2:length(res0),res0(1:2:end),'ko-',...
    1:2:length(res4),res4(1:2:end),'ks-',...
    1:4:length(res5),res5(1:4:end),'kd-');
legend('BiCGstab','GMRES','GMRES(20)')
xlabel('Number of MatVec operations');
ylabel('Norm of the Residual');

%% PRODUCE TABLE WITH NUMBER OF MATVECS:

tol = 1e-6;
OPTIONS.Tol = tol;        
OPTIONS.Disp = 0;
OPTIONS.ell = 2;

K = 0.5;
nu = 2;

for N = [10,50,150,200]
    M = N;

    f = @(x,y) sin(x).*cos(y);

    x = linspace(0,1,N);
    y = linspace(0,1,M);
    [X,Y] = meshgrid(x,y);
    eN = ones(N,1); hx = 1/(N-1);
    eM = ones(M,1); hy = 1/(M-1);
    Tx = spdiags([-eN,2*eN,-eN]./hx,-1:1,N,N);
    Ty = spdiags([-eM,2*eM,-eM]./hy,-1:1,M,M);
    Bx = spdiags([-eN,eN]./(2*hx),[-1,1],N,N);
    By = spdiags([-eM,eM]./(2*hy),[-1,1],M,M);

    A = kron(K*Tx + nu*Bx,speye(M))+ kron(speye(N),K*Ty + nu*By);
    b = reshape(f(X,Y),N*M,1);
    n = N*M;
    OPTIONS.MaxIt = n;
    
    tic; [x0,~,~,iter0,res0] = bicgstab(A,b,tol,n); time0 = toc;
    tic; [x1,hist1] = cgstab(A,b,[],[],OPTIONS); time1 = toc;
    tic; [x3,~,~,iter3,res3] = qmr(A,b,tol,n); time3 = toc;
    tic; [x4,~,~,iter4,res4] = gmres(A,b,[],tol,n); time4 = toc;
    tic; [x5,~,~,iter5,res5] = gmres(A,b,20,tol,n); time5 = toc;
    tic; [x6,~,~,iter6,res6] = bicg(A,b,tol,n); time6 = toc;
    tic; [x7,~,~,iter7,res7] = cgs(A,b,tol,n); time7 = toc;
    fprintf('%d & %d & %1.4f & %d & %1.4f & %d & %1.4f & %d & %1.4f & %d & %1.4f & %d & %1.4f & %d & %1.4f \\\\\n',...
        N,2*length(res0),time0,4*size(hist1,1),time1,length(res3),time3,length(res4),time4,length(res5),time5,2*length(res6),time6,2*length(res7),time7);
end