%**********************************************************************;
% Project           : Iterative methods and preconditioning for large 
%			and sparse linear systems with applications
%
% Program name      : ex_fixedpointprec.m
%
% Author            : F. Durastante -- fdurastante@uninsubria.it
%
% Date created      : 28 July 2017
%
% Purpose           : Example with fixed point preconditioners
%
% Revision History  :
%
% Date        Author      	Ref    Revision (Date in DD/MM/YYYY format) 
% 28/07/2017  F. Durastante     1      File Created
%
%**********************************************************************;

%% FIXED-POINT (SPLITTING METHOD) PRECONDITIONER
% In this example we try the use of Jacobi and SSOR method (with optimal
% omega coefficient) as preconditioner for the PCG method.
clear;
clc;
%% Loading the problem
load('sherman1');
A = -Problem.A;
b = -Problem.b;
n = size(A,1);
%% Unpreconditioned PCG
[~,flag0,~,iter0,resvec0] = pcg(A,b,1e-6,n);
%% Jacobi Preconditioner
M = spdiags(diag(A),0,n,n);
[~,flag1,~,iter1,resvec1] = pcg(A,b,1e-6,n,M);
%% SSOR preconditioner
% We first compute the optimal relaxtion parameter by brute force:
D = spdiags(diag(A),0,n,n);
Dinv = spdiags(1./sqrt(diag(A)),0,n,n);
E = -sparse(tril(A,-1));
F = -sparse(triu(A,1));
omegavec = 0:0.1:2;
rho = zeros(length(omegavec),1);
for i=1:length(omegavec),
    G = speye(n) - ((D-omegavec(i)*E)*(D\(D-omegavec(i)*F)))\A;
    rho(i) = max(abs(eigs(G)));
end
figure(1)
[opt,ind] = min(rho);
plot(omegavec,rho,omegavec(ind),opt,'ro');
omega = omegavec(ind);
M1 = (D - omega*E)*Dinv;
M2 = Dinv*(D - omega*F);
[~,flag3,~,iter2,resvec2] = pcg(A,b,1e-6,n,M1,M2);
%% Plot of the results
figure(2)
semilogy(0:5:iter0,resvec0(1:5:end),'k*-',... %% Unpreconditioned CG
    0:5:iter1,resvec1(1:5:end),'ko--',... %% Jacobi Preconditioner
    0:5:iter2,resvec2(1:5:end),'kx-.'); %% SSOR Preconditioner
xlabel('Iteration','interpreter','LaTeX');
ylabel('Residual','interpreter','LaTeX');
legend('CG','Jacobi Preconditioner','SSOR Preconditioner');

