/**************************************************************************
 * Matlab CMEX-file
 * Developed by K. Kaya, Johannes Langguth and B. Ucar
 *
 * Usage:
 * [match] = mmaker(A)
 * [match] = mmaker(A, matchID)
 * [match] = mmaker(A, matchID, cheapID)
 * [match] = mmaker(A, matchID, cheapID, relabel_period)
 *
 * INPUT ARGUMENTS
 * A: a sparse matrix
 * matchID: id of match algo (1-9)
 * 		1: DFS based
 *		2: BFS based
 *		3: MC21 (DFS + lookahead)
 *		4: PF (Pothen and Fan' algorithm)
 *		5: PF+ (PF + fairness)
 *		6: HK (Hopcroft and Karp's algorithm)
 *		7: HK-DW (Duff-Wiberg implementation of HK)
 *		8: ABMP (Alt et al.'s algorithm)
 *		9: ABMP-BFS (ABMP + BFS)
 *	   10: PR-FIFO-FAIR (DEFAULT)
 *
 * cheapID: id of cheap algo (1-4)
 *		1: Simple Greedy
 *		2: Karp-Sipser
 *		3: Random Karp-Sipser (DEFAULT)
 *		4: Minimum Degree (two-sided)
 *
 * relabel_period: used only when matchID = 10. Otherwise it is ignored.
 *      For the PR based algorithm, a global relabeling is started after
 *      every (m+n) x 'relabel_period' pushes where m and
 * 		n are the number of rows and columns of the matrix. Default is 1.
 * 			-1: for a global relabeling after every m pushes
 * 			-2: for a global relabeling after every n pushes
 * 		Other than these two, non-positive values are not allowed.
 *
 * OUTPUT ARGUMENTS
 * match: match of rows match(r) = c for r = 1,...,m;
 *        if c = -1, then row r is not matched; else column c and row r
 *        are matched.
 *
 * Please see the reports:
 *
 * 	"I. S. Duff, K. Kaya and B. Ucar.
 * 	'Design, Implementations and Analysis of Maximum Transversal Algorithms'
 * 	CERFACS Tech. Report TR/PA/10/76, October, 2010."
 *
 * 	"K. Kaya, J. Langguth, F. Manne and B. Ucar.
 * 	'Experiments on Push-Relabel-based Maximum Cardinality Matching Algorithms for Bipartite Graphs'
 * 	CERFACS Tech. Report TR/PA/11/33, May, 2011."
 *
 * for more details and cite them if you use the codes.
 **************************************************************************/
#include <stdio.h>
#include <math.h>
#include "mex.h"
#include "matchmaker.h"

/*inputs*/
#define A_IN          ( prhs[0] )
#define MATCH_IN      ( prhs[1] )
#define CHEAP_IN      ( prhs[2] )
#define RELABEL_IN    ( prhs[3] )
/*outputs*/
#define MATCH_OUT     ( plhs[0] )

/*...should have the following two lines for matlab version <= 7.1*/
/*
 * typedef int mwIndex;
 * typedef int mwSize;
 */

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
	int i, j, nz;

	mwSize _m, _n, _nz;
	mwIndex *irn_in;
	mwIndex *jcn_in;

	int m, n;
	int *match;
	int *row_match;
	int matchID, cheapID;
	double relabel_period;
	int* col_ptrs;
	int* col_ids;
	double *match_pr;

	relabel_period = 1;

	if (!(nrhs == 1 || nrhs == 2 || nrhs == 3  || nrhs == 4) || nlhs != 1) {
		mexErrMsgTxt("[match] = mmaker(A, matchID, cheapID)");
	}

	if (!mxIsSparse(A_IN)) {
		mexErrMsgTxt("First parameter must be a sparse matrix");
	}

	if(nrhs > 1) {
		matchID = (int) ceil(*(mxGetPr(MATCH_IN)));

		if(matchID != *(mxGetPr(MATCH_IN))) {
			mexErrMsgTxt("mmaker: matchID should be an integer\n");
		}

		if(matchID > 10 || matchID < 1) {
			mexErrMsgTxt("mmaker: matchID should be between 1 and 10\n");
		}
	} else {
		matchID = 10;
	}

	if(nrhs > 2) {
		cheapID = (int) ceil(*(mxGetPr(CHEAP_IN)));

		if(cheapID != *(mxGetPr(CHEAP_IN))) {
			mexErrMsgTxt("mmaker: cheapID should be an integer\n");
		}

		if(cheapID > 4 || cheapID < 1) {
			mexErrMsgTxt("mmaker: cheapID should be between 1 and 4\n");
		}
	} else {
		cheapID = 3;
	}

	if(nrhs > 3 && matchID == 10) {
		relabel_period = *(mxGetPr(RELABEL_IN));
		if(relabel_period <= 0 && relabel_period != -1 && relabel_period != -2) {
			mexErrMsgTxt("mmaker: relabel_period should be positive or, it should be either -1 or -2\n");
		}
	}

	_n = mxGetN(A_IN);
	_m = mxGetM(A_IN);
	jcn_in = mxGetJc(A_IN);
	irn_in = mxGetIr(A_IN);
	_nz = jcn_in[_n];

	n = (int)_n; m = (int) _m; nz = (int) _nz;
	if(n != _n || m != _m || nz != _nz) {
		mexErrMsgTxt("mmaker: problems with conversion...will abort.");
	}

	col_ptrs =  (int *) mxCalloc(sizeof(int), (n+1));
	col_ids =(int *) mxCalloc(sizeof(int), (nz));
	for(j = 0; j<=n; j++) {
		col_ptrs[j] = (int) jcn_in[j];
	}
	for(j = 0; j<nz; j++) {
		col_ids[j] = (int) irn_in[j];
	}

	match = (int*) mxCalloc(sizeof(int), n);
	row_match = (int*) mxCalloc(sizeof(int), m);

	matching(col_ptrs, col_ids, match, row_match, n, m, matchID, cheapID, relabel_period);

	MATCH_OUT = mxCreateDoubleMatrix(m, 1, mxREAL);
	match_pr  = mxGetPr(MATCH_OUT);
	for (i = 0; i < m; i++) {
		if(row_match[i] == -1) {
			match_pr[i] = -1;
		} else {
			match_pr[i] = row_match[i] + 1;
		}
	}

	mxFree(row_match);
	mxFree(match);
	mxFree(col_ptrs);
	mxFree(col_ids);
	return;
}



