%**********************************************************************;
% Project           : Iterative methods and preconditioning for large 
%			and sparse linear systems with applications
%
% Program name      : ex_helmoltz.m
%
% Author            : F. Durastante -- fdurastante@uninsubria.it
%
% Date created      : 28 July 2017
%
% Purpose           : Preconditioners update for Helmoltz equation
%
% Revision History  :
%
% Date        Author      	Ref    Revision (Date in DD/MM/YYYY format) 
% 28/07/2017  F. Durastante     1      File Created
%
%**********************************************************************;
%% Helmoltz Equation Test

clear;
clc;

addpath('../chapter3/');

fprintf('Input the number of discretization nodes.\n')
N = input('n = ');
fprintf('Input the number of step s.\n')
s = input('s = ');
fprintf('Choice of the diffusion coefficient:\n1) c(x,y) = 1\n2) c(x,y) = exp(-x-y)\n');
choice = input('c (1/2) = ');
if choice == 1
    aa = @(x,y) 1;
elseif choice == 2
    aa = @(x,y) exp(-x-y);
else
    error('Coefficient unrecognized!')
end

% Construction of the matrix of the operator:
x_max=1;
dx=x_max/(N+1);
d2=1/(dx*dx);
N2=N*N;
H=spalloc(N2,N2,5*N2-4*N);
for j1=0:N-1
    for j2=1:N
        rr=j1*N+j2; % natural ordering
        j1p=j1+1; ru=j1p*dx;
        aa1=aa((j2-.5)*dx,j1p*dx)*d2; aa2=aa((j2+.5)*dx,j1p*dx)*d2;
        aa3=aa(j2*dx,(j1p+.5)*dx)*d2; aa4=aa(j2*dx,(j1p-.5)*dx)*d2;
        H(rr,rr)=-aa1-aa2-aa3-aa4;
        if (j2>1),   H(rr,rr-1)=aa1; end
        if (j2<N),   H(rr,rr+1)=aa2; end
        if (j1<N-1), H(rr,rr+N)=aa3; end
        if (j1>0),   H(rr,rr-N)=aa4; end
    end %j2
end %j1
H=-H;

% Solution step
ind = 1;
droptol=1e-1;
k = 1;                                  % Degree of the approximation
fprintf('$\\sigma_1$ & IT & T(s) & IT & T(s) \\\\ \n');
for sigma1 = [50,100,200,400,800]
    A0 = H/d2;% + sigma1*speye(N2,N2)/d2;
    [Z,p] = ainv(A0,droptol); % Reference preconditioner
    Zt = Z';
    D = spdiags(p,0,N2,N2);
    for j = 1:s+1
        sigma2mat = 1i*spdiags(1000*rand(N2,1),0,N2,N2)/d2;
        % Correction Term:
        k=0;        % Diagonal Correction Term
        E1 = spdiags(spdiags(Zt*(imag(sigma2mat))*Z,-k:k),-k:k,N2,N2);
        P1 = @(x) Z*((D + E1)\(Zt*x));
        k=1;        % Tridiagonal Correction Term
        E2 = spdiags(spdiags(Zt*(imag(sigma2mat))*Z,-k:k),-k:k,N2,N2);
        P2 = @(x) Z*((D + E2)\(Zt*x));
        k=2;        % Pentadiagonal Correction Term
        E3 = spdiags(spdiags(Zt*(imag(sigma2mat))*Z,-k:k),-k:k,N2,N2);
        P3 = @(x) Z*((D + E3)\(Zt*x));
        tic;
        [X0,FLAG0,RELRES0,ITER0(:,j)] = gmres(A0+sigma2mat,rand(N2,1)/d2,[],1e-6,N2);
        time0(j) = toc;
        tic;
        [X1,FLAG1,RELRES1,ITER1(:,j)] = gmres(A0+sigma2mat,rand(N2,1)/d2,[],1e-6,N2,@(x) P1(x));
        time1(j) = toc;
        tic;
        [X2,FLAG2,RELRES2,ITER2(:,j)] = gmres(A0+sigma2mat,rand(N2,1)/d2,[],1e-6,N2,@(x) P2(x));
        time2(j) = toc;
        tic;
        [X3,FLAG3,RELRES3,ITER3(:,j)] = gmres(A0+sigma2mat,rand(N2,1)/d2,[],1e-6,N2,@(x) P3(x));
        time3(j) = toc;
    end
    ITER0m(ind) = mean(ITER0(2,:));
    TIME0m(ind) = mean(time0);
    ITER1m(ind) = mean(ITER1(2,:));
    TIME1m(ind) = mean(time1);
    ITER2m(ind) = mean(ITER2(2,:));
    TIME2m(ind) = mean(time2);
    ITER3m(ind) = mean(ITER3(2,:));
    TIME3m(ind) = mean(time3);
    fprintf('%d & %d & %1.2e & %d & %1.2e & %d & %1.2e & %d & %1.2e \\\\\n',...
        sigma1,round(ITER0m(ind)),TIME0m(ind)...
        ,round(ITER1m(ind)),TIME1m(ind)...
        ,round(ITER2m(ind)),TIME2m(ind)...
        ,round(ITER2m(ind)),TIME2m(ind));
    ind = ind+1;
end

