%% SOLUTION OF TOEPLITZ LINEAR SYSTEM WITH KERNEL PRECONDITIONER
% We test the solution of some Toeplitz linear system with circulant
% preconditioner generated with the Kernel approach. This experiments are
% relative to the case in which we know the symbol/generating function of
% the sequence of Toeplitz matrices

clear;
clc;

%% Experiment 1
% We consider the function f(t) = t^4 + 1, t \in [-\pi,\pi].

n = 512;
t = zeros(n,1);
t(1) = 1+(pi^4)/5;
for j=2:n
    t(j) = (4*(-1)^(j-1))*(-6 + (j-1)^2*pi^2)/((j-1)^4);
end
A = toeplitz(t);
b = ones(n,1);

figure(1);
subplot(1,2,1);
x = linspace(0,pi,n);
plot(x,x.^4+1,'k-',x,eig(A),'ko');
title('Eigenvalue Distribution','Interpreter','LaTeX');
legend('Symbol','Eigenvalues')

T = @(x) toepmv(A,x);

% Solution paramaters
tol = 1e-7;
maxit = n;

% Reference solution
fprintf('Non preconditioned:\n');
x = pcg(T,b,tol,maxit);
% T. Chan Preconditioner
fprintf('\nTony Chan Preconditioner:\n');
cchan = gencircprec( t,'tchan' );
Cchan = @(x) cprec(cchan,x);
x = pcg(T,b,tol,maxit,@(x) Cchan(x));
% Strang Preconditioner
fprintf('\nStrang Preconditioner:\n');
cstrang = gencircprec( t,'strang');
Cstrang = @(x) cprec(cstrang,x);
x = pcg(T,b,tol,maxit,@(x) Cstrang(x));
% R. Chan Preconditioner
fprintf('\nRaymond Chan Preconditioner:\n');
crchan = gencircprec( t,'rchan');
Crchan = @(x) cprec(crchan,x);
x = pcg(T,b,tol,maxit,@(x) Crchan(x));
% Modified Dirichlet Preconditioner
fprintf('\nModified Dirichlet Preconditioner:\n');
cdirichlet = gencircprec( t,'dirichlet');
Cdirichlet = @(x) cprec(cdirichlet,x);
x = pcg(T,b,tol,maxit,@(x) Cdirichlet(x));
% de la Vallee Poussin Preconditioner
fprintf('\nde la Vallee Poussin Preconditioner:\n');
cpoussin = gencircprec( t,'poussin');
Cpoussin = @(x) cprec(cpoussin,x);
x = pcg(T,b,tol,maxit,@(x) Cpoussin(x));
% von Hann Preconditioner
fprintf('\nvon Hann Preconditioner:\n');
chann = gencircprec( t,'hann');
Chann = @(x) cprec(chann,x);
x = pcg(T,b,tol,maxit,@(x) Chann(x));
% Hamming preconditioner
fprintf('\nHamming Preconditioner:\n');
chamming = gencircprec( t,'hamming');
Chamming = @(x) cprec(chamming,x);
x = pcg(T,b,tol,maxit,@(x) Chamming(x));
% Bernstein preconditioner
fprintf('\nBernstein Preconditioner:\n');
cbernstein = gencircprec( t,'bernstein');
Cbernstein = @(x) cprec(cbernstein,x);
x = pcg(T,b,tol,maxit,@(x) Cbernstein(x));
% Superoptimal preconditioner
fprintf('\nSuperoptimal Preconditioner:\n');
csuperoptimal = gencircprec( t,'superoptimal');
Csuperoptimal = @(x) cprec(csuperoptimal,x);
x = pcg(T,b,tol,maxit,@(x) Csuperoptimal(x));

figure(2)
semilogx(eig(A),0,'kx',eig(Cchan(A)),0.1,'kx',eig(Cstrang(A)),0.2,'kx',...
    eig(Crchan(A)),0.3,'kx',eig(Cdirichlet(A)),0.4,'kx',...
    eig(Cpoussin(A)),0.5,'kx',eig(Chann(A)),0.6,'kx',....
    eig(Chamming(A)),0.7,'kx',real(eig(Cbernstein(A))),0.8,'kx',...
    real(eig(Csuperoptimal(A))),0.9,'kx');
axis([1e-1 pi^4+1 -0.1 1])
yticks(0:0.1:1)
yticklabels({'Non preconditioned','T. Chan','Strang','R. Chan',...
    'Modified Dirichlet','de la Vallee Poussin','von Hann','Hamming',...
    'Bernstein','Super Optimal'})
set(0,'DefaultTextInterpreter', 'latex')
line([1 1], get(gca, 'ylim'),'color','k','LineStyle','--');

%% Experiment 2
% We consider the function f(t) = t^4 + t^2 + 1, t \in [-\pi,\pi].

n = 128;
t = zeros(n,1);
t(1) = 1 + pi^(2)/3 + pi^(4)/5;
for j=2:n
    t(j) = (2*(-1)^(j-1))*(-12 + (j-1)^2*(1+2*pi^2))/((j-1)^4);
end
A = toeplitz(t);
b = ones(n,1);

figure(1);
subplot(1,2,2);
x = linspace(0,pi,n);
plot(x,x.^4+x.^2+1,'k-',x,eig(A),'ko');
title('Eigenvalue Distribution','Interpreter','LaTeX');
legend('Symbol','Eigenvalues')

T = @(x) toepmv(A,x);

% Solution paramaters
tol = 1e-7;
maxit = n;

% Reference solution
fprintf('Non preconditioned:\n');
x = pcg(T,b,tol,maxit);
% T. Chan Preconditioner
fprintf('\nTony Chan Preconditioner:\n');
cchan = gencircprec( t,'tchan' );
Cchan = @(x) cprec(cchan,x);
x = pcg(T,b,tol,maxit,@(x) Cchan(x));
% Strang Preconditioner
fprintf('\nStrang Preconditioner:\n');
cstrang = gencircprec( t,'strang');
Cstrang = @(x) cprec(cstrang,x);
x = pcg(T,b,tol,maxit,@(x) Cstrang(x));
% R. Chan Preconditioner
fprintf('\nRaymond Chan Preconditioner:\n');
crchan = gencircprec( t,'rchan');
Crchan = @(x) cprec(crchan,x);
x = pcg(T,b,tol,maxit,@(x) Crchan(x));
% Modified Dirichlet Preconditioner
fprintf('\nModified Dirichlet Preconditioner:\n');
cdirichlet = gencircprec( t,'dirichlet');
Cdirichlet = @(x) cprec(cdirichlet,x);
x = pcg(T,b,tol,maxit,@(x) Cdirichlet(x));
% de la Vallee Poussin Preconditioner
fprintf('\nde la Vallee Poussin Preconditioner:\n');
cpoussin = gencircprec( t,'poussin');
Cpoussin = @(x) cprec(cpoussin,x);
x = pcg(T,b,tol,maxit,@(x) Cpoussin(x));
% von Hann Preconditioner
fprintf('\nvon Hann Preconditioner:\n');
chann = gencircprec( t,'hann');
Chann = @(x) cprec(chann,x);
x = pcg(T,b,tol,maxit,@(x) Chann(x));
% Hamming preconditioner
fprintf('\nHamming Preconditioner:\n');
chamming = gencircprec( t,'hamming');
Chamming = @(x) cprec(chamming,x);
x = pcg(T,b,tol,maxit,@(x) Chamming(x));
% Bernstein preconditioner
fprintf('\nBernstein Preconditioner:\n');
cbernstein = gencircprec( t,'bernstein');
Cbernstein = @(x) cprec(cbernstein,x);
x = pcg(T,b,tol,maxit,@(x) Cbernstein(x));
% Superoptimal preconditioner
fprintf('\nSuperoptimal Preconditioner:\n');
csuperoptimal = gencircprec( t,'superoptimal');
Csuperoptimal = @(x) cprec(csuperoptimal,x);
x = pcg(T,b,tol,maxit,@(x) Csuperoptimal(x));


figure(3)
semilogx(eig(A),0,'kx',eig(Cchan(A)),0.1,'kx',eig(Cstrang(A)),0.2,'kx',...
    eig(Crchan(A)),0.3,'kx',eig(Cdirichlet(A)),0.4,'kx',...
    eig(Cpoussin(A)),0.5,'kx',eig(Chann(A)),0.6,'kx',....
    eig(Chamming(A)),0.7,'kx',real(eig(Cbernstein(A))),0.8,'kx',...
    real(eig(Csuperoptimal(A))),0.9,'kx');
axis([1e-1 pi^4+1 -0.1 1])
yticks(0:0.1:1)
yticklabels({'Non preconditioned','T. Chan','Strang','R. Chan',...
    'Modified Dirichlet','de la Vallee Poussin','von Hann','Hamming',...
    'Bernstein','Super Optimal'})
set(0,'DefaultTextInterpreter', 'latex')
line([1 1], get(gca, 'ylim'),'color','k','LineStyle','--');

