%% TOEPLITZ PRECONDITIONERS AND TOEPLITZ APPROXIMATE INVERSE PRECONDITIONERS
% This code contains some example relative to the section of the book
% called Toeplitz and band--Toeplitz preconditioner, for some routines it
% depends on the code from the section on circulant preconditioners.

clear;
clc;

%% EXAMPLE 1: The Hanke e Nagy preconditioner
n = 2^10;
e = ones(n,1);
T = spdiags([-e/4,-e/4,e,-e/4,-e/4,],[-6,-1,0,1,6],n,n);
figure(1)
spy(T);
% Building the Hanke--Nagy preconditioner (only the first column is needed)
beta = bandwidth(T)+1;
c = [T(1:n,1);T(1,beta:-1:2)'];
Lambda = real(fft(full(c)));
Lambdainv = zeros(size(Lambda)); % Compute the inverse element removing the
                                 % zero entries
for i=1:length(Lambda)
    if Lambda(i) > 0
        Lambdainv(i) = 1/Lambda(i);
    end
end
P = @(x) hankenagy( Lambdainv,x,n,beta );
% T. Chan Preconditioner:
cchan = gencircprec( full(T(:,1)),'tchan' );
Cchan = @(x) cprec(cchan,x);
% right-hand side:
b = T*ones(n,1);
tol = 1e-7;
maxit = n;
% Fast Toeplitz product
n = size(T,1); 
t = [T(:,1);0;T(1,n:-1:2)'];
t = fft(full(t));

x1 = pcg(@(x) toepmv(full(T),x),b,tol,maxit); % Without preconditioner
x2 = pcg(@(x) toepmv(t,x),b,tol,maxit,@(x) P(x)); % Hanke & Nagy preconditioner
x3 = pcg(@(x) toepmv(full(T),x),b,tol,maxit,@(x) Cchan(x)); % Tony Chan preconditioner

% Plotting the Cluster: (use this routines only for small dimensions, it
% can become taxing on the memory!)
for i=1:n
   H(:,i)=P(full(T(:,i))); 
end
figure(2)
semilogx(eig(T),0,'kx',eig(Cchan(full(T))),0.1,'kx',real(eig(H)),0.2,'kx');
axis([1e-1 10 -0.05 0.25])
yticks(0:0.1:1)
yticklabels({'Non preconditioned','T. Chan','Hanke Nagy'})
set(0,'DefaultTextInterpreter', 'latex')
line([1 1], get(gca, 'ylim'),'color','k','LineStyle','--');

%% EXAMPLE 2: The Chan e Ng preconditioner
clear;
clc;

n = 2^8;
e = ones(n,1);
t = toepgenerator(n,4);

s = 3;
ts = toepgenerator(n*s,4);
c = gencircprec(ts','tchan');
for k = 0:n-1
    zp(k+1) = 1/(s*n)*(sum((1./c).*exp(-2*pi*1i*((0:(s*n-1))')*k/(s*n))));
    zm(k+1) = 1/(s*n)*(sum((1./c).*exp(-2*pi*1i*((0:(s*n-1))')*(-k)/(s*n))));
end

% Fast Toeplitz product
T = toeplitz(t);
n = size(T,1); 
t = [T(:,1);0;T(1,n:-1:2)'];
t = fft(full(t));
% Fast Toeplitz product preconditioner
Ps = toeplitz(zp,zm);
n = size(T,1); 
p = [Ps(:,1);0;Ps(1,n:-1:2)'];
p = fft(full(p));
% T. Chan Preconditioner:
cchan = gencircprec( full(T(:,1)),'tchan' );
Cchan = @(x) cprec(cchan,x);

b = ones(n,1);
tol = 1e-7;
maxit = n;

x1 = pcg(@(x) toepmv(t,x),b,tol,maxit);
x2 = pcg(@(x) toepmv(t,x),b,tol,maxit,@(x) toepmv(p,x));
x3 = pcg(@(x) toepmv(t,x),b,tol,maxit,@(x) Cchan(x));

% Plotting the Cluster: (use this routines only for small dimensions, it
% can become taxing on the memory!)
for i=1:n
   H(:,i)=toepmv(p,T(:,i));
end
figure(2)
semilogx(eig(T),0,'kx',eig(Cchan(full(T))),0.1,'kx',real(eig(H)),0.2,'kx');
axis([1e-1 max(max(eig(T))) -0.1 0.3])
yticks(0:0.1:1)
yticklabels({'Non preconditioned','T. Chan','P_n^{(s)}'})
set(0,'DefaultTextInterpreter', 'latex')
line([1 1], get(gca, 'ylim'),'color','k','LineStyle','--');
set(0,'DefaultTextInterpreter', 'LaTeX')

%% Example 3
clear;
clc;

for zz=13
n = 2^zz;                        % size of the problem
t = toepgenerator(n,8);         % problem: f(\theta) = \theta^4 (1)
t2 = toepgenerator(n,6);        % problem: f(\theta) = \theta^2 (2)
T = toeplitz(t);
T2 = toeplitz(t2);
l = 2;                          % half of the order of the zero (1):
tp = zeros(n,1);
for k=0:l
    tp(k+1,1) = (-1)^k*nchoosek(2*l,l+k);
end
P = sparse(toeplitz(tp));
l = 1;                          % half of the order of the zero (2):
tp = zeros(n,1);
for k=0:l
    tp(k+1,1) = (-1)^k*nchoosek(2*l,l+k);
end
P2 = sparse(toeplitz(tp));

% Jackson Kernel preconditioner:
c12 = jacksonprec(t,2); J12 = @(x) cprec(c12,x);
c13 = jacksonprec(t,3); J13 = @(x) cprec(c13,x);
c14 = jacksonprec(t,4); J14 = @(x) cprec(c14,x);
c22 = jacksonprec(t2,2); J22 = @(x) cprec(c22,x);
c23 = jacksonprec(t2,3); J23 = @(x) cprec(c23,x);
c24 = jacksonprec(t2,4); J24 = @(x) cprec(c24,x);

% Fast product for Toeplitz (1)
n = size(T,1);                       % Compute the circulant extension for fast
t = [T(:,1);0;conj(T(1,n:-1:2))'];    % products
t = fft(t);
% Fast product for Toeplitz (2)
n = size(T2,1);                       % Compute the circulant extension for fast
t2 = [T2(:,1);0;conj(T2(1,n:-1:2))'];    % products
t2 = fft(t2);
% Data:
b = toepmv(t,ones(n,1));
tol = 1e-7;
maxit = 1000;

% Solution for f(\theta) = \theta^4
[x11,~,~,iter11] = pcg(@(x) toepmv(t,x),b,tol,maxit);
[x21,~,~,iter21] = pcg(@(x) toepmv(t,x),b,tol,maxit,P);
[x31,~,~,iter31] = pcg(@(x) toepmv(t,x),b,tol,maxit,J12);
[x41,~,~,iter41] = pcg(@(x) toepmv(t,x),b,tol,maxit,J13);
[x51,~,~,iter51] = pcg(@(x) toepmv(t,x),b,tol,maxit,J14);
% Solution for f(\theta) = \theta^2
[x12,~,~,iter12] = pcg(@(x) toepmv(t2,x),b,tol,maxit);
[x22,~,~,iter22] = pcg(@(x) toepmv(t2,x),b,tol,maxit,P2);
[x32,~,~,iter32] = pcg(@(x) toepmv(t2,x),b,tol,maxit,J22);
[x42,~,~,iter42] = pcg(@(x) toepmv(t2,x),b,tol,maxit,J23);
[x52,~,~,iter52] = pcg(@(x) toepmv(t2,x),b,tol,maxit,J24);

fprintf('%d & %d & %d & %d & %d & %d \n',...
    n,iter11,iter21,iter31,iter41,iter51);
fprintf('%d & %d & %d & %d & %d & %d \n',...
    n,iter12,iter22,iter32,iter42,iter52);
    
end
% Plotting the Cluster: (use this routines only for small dimensions, it
% can become taxing on the memory!)
% figure(4)
% semilogx(eig(T),0,'kx',eig(P\T),0.1,'kx',eig(T2),0.2,'kx',...
%     eig(P2\T2),0.3,'kx',eig(J22(T2)),0.4,'kx',...
%     eig(J23(T2)),0.5,'kx',eig(J24(T2)),0.6,'kx');
% axis([min(eig(T)) max(eig(T)) -0.1 0.8])
% yticks(0:0.1:1)
% yticklabels({'f(x) = x^4','P(s_2)^{-1}T_n(f)','f(x) = x^2','P(s_1)^{-1}T_n(f)'})
% set(0,'DefaultTextInterpreter', 'latex')
% line([1 1], get(gca, 'ylim'),'color','k','LineStyle','--');
% set(0,'DefaultTextInterpreter', 'LaTeX')
